#!/usr/bin/env python
'''
Export to ROOT TGeo
'''

import ROOT

from gegede.util import wash_units

def Amass(obj):
    return obj.a.to('g/mole').magnitude

def Nnuc(obj):
    return int(round(obj.a.to('g/mole').magnitude))

def Symbol(obj):
    return obj.symbol.upper()

# Note, ROOT density is hard coded as g/cc
#  http://root.cern.ch/root/html534/guides/users-guide/Geometry.html#units
def Density(obj):
    return obj.density.to('g/cc').magnitude


def get_element(tgeo, name):
    et = tgeo.GetElementTable()
    print et
    ele = et.FindElement(name)

    if not ele:
        nele = et.GetNelements()
        print 'I know about %d elements:' % nele
        for count in range(nele):
            known = et.GetElement(count)
            print '\t%4d: "%s" "%s"' % (count, known.GetName(), known.GetTitle())
        raise KeyError, 'No such element: "%s"' % name
    return ele
    

# fixme, need to handle material properties with obj.params
def make_material(bucket, mat):
    '''
    Make and return a TGeo equivalent to the gegede obj.
    '''
    typename = type(mat).__name__
    new = None

    if typename == 'Element':
        new = bucket.make(ROOT.TGeoElement, mat.name, Symbol(mat), mat.z, Amass(mat))

    if typename == 'Isotope':
        new = bucket.make(ROOT.TGeoIsotope, mat.name, mat.z, Nnuc(mat), Amass(mat))

    if typename == 'Composition': # element mix of isotopes
        new = bucket.make(ROOT.TGeoElement, mat.name, Symbol(mat), len(mat.isotopes))
        for count, (isoname, isofrac) in enumerate(mat.isotopes):
            iso = bucket.get(ROOT.TGeoIsotope, isoname)
            assert iso, 'No isotope named "%s"' % isoname
            new.AddIsotope(iso, isofrac)

    if typename == 'Amalgam':
        new = bucket.make(ROOT.TGeoMaterial, mat.name, Amass(mat), float(mat.z), Density(mat))

    if typename == 'Molecule':  # mix of elements
        new = bucket.make(ROOT.TGeoMixture, mat.name, len(mat.elements), Density(mat))
        for elename, elenum in mat.elements:
            ele = bucket.get(ROOT.TGeoElement, elename)
            assert ele, 'No element named "%s"' % elename
            new.AddElement(ele, elenum)

    if typename == 'Mixture':
        new = bucket.make(ROOT.TGeoMixture, mat.name, len(mat.components), Density(mat))
        for compname, compfrac in mat.components:
            comp = bucket.get(ROOT.TGeoMaterial, compname) or bucket.get(ROOT.TGeoElement, compname)
            assert comp, 'No mat/ele named "%s"' % compname
            new.AddElement(comp, compfrac)

    if not new:
        raise ValueError, 'Unknown type: "%s"' % typename

    bucket.add(new, 'MATERIAL')
    return new

def make_shape(bucket, shape):
    '''
    Make a TGeo shape out of the object
    '''
    typename = type(shape).__name__
    shape, _ = wash_units(shape)

    obj = None
    if typename == 'Box':
        obj = bucket.make(ROOT.TGeoBBox, shape.name, 
                          shape.dx, shape.dy, shape.dz)

    if typename == 'Tubs':
        obj = bucket.make(ROOT.TGeoTubeSeg, shape.name,
                          shape.rmin, shape.rmax, shape.dz,
                          shape.sphi, shape.sphi+shape.dphi)
    
    if typename == 'Sphere':
        obj = bucket.make(ROOT.TGeoSphere, shape.name,
                          shape.rmin, shape.rmax,
                          shape.stheta, shape.stheta+shape.dtheta,
                          shape.sphi, shape.sphi+shape.dphi)

    # etc....  Grow this as gegede.schema.Schema.shapes grows....

    if obj:                     # for generic look up later
        bucket.add(obj, 'SHAPE')

    return obj


def make_medium(bucket, matname):
    medname = matname+'_medium'
    med = bucket.get('MEDIUM', medname)
    if med: return med

    mat = bucket.get('MATERIAL', matname)
    if not mat: return None

    med = bucket.make(ROOT.TGeoMedium, medname, len(bucket.store('MEDIUM')), mat)
    bucket.add(med, 'MEDIUM')
    return med

def make_structure(bucket, obj):
    typename = type(obj).__name__

    if typename == 'Volume':
        shape = bucket.get('SHAPE', obj.shape)
        assert shape, 'No shape "%s"' % obj.shape
        med = make_medium(bucket, obj.material) 
        assert med, 'No medium for material "%s' % obj.material
        vol = bucket.make(ROOT.TGeoVolume, obj.name, shape, med)
        
                          


    return


    

class Bucket(object):
    def __init__(self, world):
        if not isinstance(world, type("")):
            world = world.name
        from collections import defaultdict
        self._dat = defaultdict(dict)
        self.tgeo = ROOT.TGeoManager(world, 'GeGeDe %s' % world)

    def store(self, category):
        return self._dat[category]

    def get(self, category, name):
        'Get object named <name> from <category> (label or type)'
        if not isinstance(category, type("")):
            category = category.__name__
        return self._dat[category].get(name)

    def has(self, obj):
        return self.get(type(obj), obj.GetName())

    def add(self, obj, category=None):
        self._dat[category][obj.GetName()] = obj

    def make(self, TYPE, name, *args):
        obj = self.get(TYPE, name)
        if obj:
            print 'Object %s %s already made' % (TYPE, name)
            return obj
        obj = TYPE(name, *args)
        ROOT.SetOwnership(obj, 0)
        self.add(obj, TYPE.__name__)
        print 'Made: %s' % obj
        return obj

    pass

def convert(geom):
    '''
    Return a ROOT.TGeoManager filled with the geometry.
    '''

    bucket = Bucket(geom.world)
    # Materials
    for obj in geom.store.matter.values():
        make_material(bucket, obj)

    # Shapes
    for obj in geom.store.shapes.values():
        make_shape(bucket, obj)

    # Structure
    for obj in geom.store.structure.values():
        make_structure(bucket, obj)

    return bucket

def validate(tgeo):
    return True

def dumps(tgeo):
    return None



    



#!/usr/bin/env python
'''
Special schema types
'''

try:
    python_string = basestring
except NameError:
    python_string = str

from .. import Quantity

def isquantity(thing):
    if isinstance(thing, Quantity):
        return True             # it's an instance

    if isinstance(thing, python_string):
        try:
            Quantity(thing)
        except ValueError:
            pass
        else:
            return True         # it's a string that can be parsed
    return False                # dunno what it is

def toquantity(proto):
    qobj = Quantity(proto)      # wash to assure Quantity object
    def converter(other):
        other = Quantity(other)
        if qobj.dimensionality == other.dimensionality:
            return other
        raise ValueError, 'Unit mismatch: %s incompatible with prototype %s' % (other, qobj)
    return converter

def Named(thing):
    '''
    Return thing.name if it has the attribute, otherwise return thing
    '''
    if hasattr(thing,'name'):
        return thing.name
    return thing


def NamedTypedList(typ, minentries = 1):
    '''Return a function which processes a prototype which is a list of
    2-tuples: ("name", obj) where <obj> is an instance of type <typ>.
    '''

    def named_typed_list_converter(proto):
        #print 'named_typed_list_converter(%s, "%s")' % (typ, str(proto))
        ret = list()
        if len(proto) < minentries:
            raise ValueError, "At least %d entries required" % minentries
        for n,v in proto:
            v = typ(v)
            n = Named(n)
            ret.append((n,v))
        return ret
    named_typed_list_converter.default = list
    return named_typed_list_converter


def NameList(typ, minentries = 1):
    '''Return a function which processes a prototype which is a list of
    object names.
    '''

    def name_list_converter(proto):
        #print 'name_list_converter(%s, "%s")' % (typ, str(proto))
        ret = list()
        if len(proto) < minentries:
            raise ValueError, "At least %d entries required" % minentries
        for n in proto:
            n = Named(n)
            ret.append(n)
        return ret
    name_list_converter.default = list
    return name_list_converter




#!/usr/bin/env python
'''Export GeGeDe to an Open Inventor scene graph.

A SG allows for a more flexible layout than is required to represent
the nested volume structure followed by GeGeDe.  A choice is made to
map this into an SG.

Bot logical volumes (LV) and placements (PV) are represented with
SoSeparators.  An LV SoSeparator may contain zero or more PVs followed
by an optional SoMaterial and a SoShape.  If an LV contains only PVs
it is an assembly.  A PV is a SoSeparator with a transform and an LV.

'''

from collections import defaultdict
from pivy import coin

spatial_units = 'mm'

def pos2trans(obj):
    pos = coin.SoTransform()
    pos.translation = (obj.x.to('mm').magnitude,
                       obj.y.to('mm').magnitude,
                       obj.z.to('mm').magnitude)
    return pos

def rot2trans(obj):
    tot = coin.SoGroup()
    for ang,axis in[(obj.x.to('radian').magnitude,coin.SoRotationXYZ.X),
                    (obj.y.to('radian').magnitude,coin.SoRotationXYZ.Y),
                    (obj.z.to('radian').magnitude,coin.SoRotationXYZ.Z)]:
        if ang == 0.0:
            continue
        r = coin.SoRotationXYZ()
        r.axis = axis
        r.angle = ang
        tot.addChild(r)
    assert len(tot), 'Failed to add any rotations for %s' % str(obj.name)
    return tot

# http://stackoverflow.com/questions/6283080/random-unit-vector-in-multi-dimensional-space
from random import gauss
def make_rand_vector(dims = 3):
    vec = [gauss(0, 1) for i in range(dims)]
    mag = sum(x**2 for x in vec) ** .5
    return [x/mag for x in vec]

def mat2mat(obj):
    '''
    Convert gegede material to SoMaterial.  They have nothing in common.

    ambientColor
    diffuseColor
    specularColor
    emissiveColor
    shininess
    transparency
    '''
    # fixme: add checking of hints
    out = coin.SoMaterial()
    out.diffuseColor = make_rand_vector()
    out.specularColor = make_rand_vector()
    out.ambientColor = make_rand_vector()
    out.transparency = 0.5
    return out

def shape2shape(name, geom):
    '''
    Convert from gegede shape to a SoShape
    '''
    obj = geom.get_shape(name)
    if type(obj).__name__ == 'Box':
        box = coin.SoCube()
        box.width  = 2.0*obj.dx.to(spatial_units).magnitude
        box.height = 2.0*obj.dy.to(spatial_units).magnitude
        box.depth  = 2.0*obj.dz.to(spatial_units).magnitude
        return box

    if type(obj).__name__ == 'Tubs':
        tub = coin.SoCylinder()
        tub.radius = obj.rmax.to(spatial_units).magnitude
        tub.height = 0.5*obj.dz.to(spatial_units).magnitude
        # fixme: no built-in support of rmin, nor start/stop angles - need to make boolean
        return tub

    if type(obj).__name__ == 'Sphere':
        sph = coin.SoSphere()
        sph.radius = obj.rmax.to(spatial_units).magnitude
        # fixme: no built-in support of rmin, nor start/stop angles - need to make boolean
        return sph

    if type(obj).__name__ == 'Boolean':
        print 'WARNING: Boolean shapes are not supported, simply using the first shape for %s' % obj.name
        return shape2shape(obj.first, geom)

    raise ValueError, 'Unsupported shape for scenegraph: "%s"' % type(obj).__name__

def convert(geom):
    '''
    Return the top node of an Open Inventor scenegraph made from the geom.
    '''

    materials = dict()
    for obj in geom.store.matter.values():
        materials[obj.name] = mat2mat(obj)
        continue

    positions = dict()
    rotations = dict()
    for obj in geom.store.structure.values():
        typename = type(obj).__name__.lower()
        if typename == 'position':
            positions[obj.name] = pos2trans(obj)
            continue
        if typename == 'rotation':
            rotations[obj.name] = rot2trans(obj)
            continue

    placements = defaultdict(coin.SoSeparator)
    volumes = defaultdict(coin.SoSeparator)
    for obj in geom.store.structure.values():
        typename = type(obj).__name__.lower()
        if typename != 'placement':
            continue
        pnode = placements[obj.name]
        if obj.pos:
            pos = positions[obj.pos]
            pnode.addChild(pos)
        if obj.rot:
            rot = rotations[obj.rot]
            pnode.addChild(rot)
        vol = volumes[obj.volume]
        pnode.addChild(vol)
    
    for obj in geom.store.structure.values():
        typename = type(obj).__name__.lower()
        if typename != 'volume':
            continue
        vol = volumes[obj.name]
        for place in obj.placements:
            vol.addChild(placements[place])
        if obj.shape:
            vol.addChild(shape2shape(obj.shape, geom))
            vol.addChild(materials[obj.material])
        
    return volumes[geom.world]

def output(top, filename):
    '''
    '''
    writer = coin.SoWriteAction()
    out = writer.getOutput()
    out.openFile(filename)
    out.setBinary(False)
    writer.apply(top)
    out.closeFile()



#!/usr/bin/env python
'''
Example gegede builders for a trivial LAr geometry
'''

import gegede.builders

class WorldBuilder(gegede.builders.Builder):
    '''
    Build a big box world volume.
    '''

    def configure(self, size='50m', material='Rock', **kwds):
        self.size = size
        self.material = material

    def construct(self, geom):
        shape = geom.shapes.Box(self.name + '_box_shape', dx=self.size, dy=self.size, dz=self.size)
        lv = geom.structure.Volume(self.name+'_volume', material=self.box_mat, shape=shape)
        self.add_volume(lv)

        site_lv = self.builders[0].volumes[0] # expect one sub-builder with one logical volume
        geom.structure.Placement('%s_in_%s' % (site_lv.name, self.name), volume = site_lv)
        return

class SiteBuilder(gegede.builders.Builder):
    def configure(self, site='homestake', material = 'Air', **kwds):
        self.site = site.lower()
        self.material = material

    def construct(self, geom):
        if self.site == 'homestake':
            shape = geom.structure.Box('homestake site box', dx='10m', dy='10m', dz='10m')
        elif self.site == '35t':
            shape = geom.structure.Box('35t site box', dx='5m', dy='5m', dz='5m')
        else:
            raise ValueError, 'Unknown site: "%s"' % self.site
        lv = geom.structure.Volume(self.site + '_volume', material=self.material, shape = self.shape)



#!/usr/bin/env python
'''
Methods to iterate over a geometry.
'''

def ascending_all(store, top):
    '''Iterate from children to parents up to given top volume.
    '''
    if isinstance(top, type("")):
        top = store[top]
    for pname in top.placements:
        place = store[pname]
        vname = place.volume
        daughter = store[vname]
        for ret in ascending_all(store, daughter):
            yield ret
    yield top
            
def ascending(store, top):
    '''Iterate from children to parents up to given top volume.
    '''
    seen = set()
    ret = list()
    for vol in ascending_all(store, top):
        if vol.name in seen:
            continue
        seen.add(vol.name)
        ret.append(vol)
    return ret

 






from collections import namedtuple, OrderedDict
import schema as default_schema
from schema.tools import make_maker
from .util import list_match

class Geometry(object):
    '''A geometry constructor.

    This object provides for creation of elements in a geometry and
    maintains a store of these elements for later retrieval by name.
    '''

    # Name of world volume, set via .set_world()
    _world = None

    def __init__(self, schema = None):
        '''Create a geometry constructor.

        The optional <schema> object is a raw schema and defaults to
        gegede.schema.Schema.
        '''
        if not schema:
            schema = default_schema.Schema

        parts = sorted(schema.keys())
        self.store = namedtuple('GeometryStore', parts)(*[OrderedDict() for x in parts])

        for part, store in zip(parts,self.store):
            scheme = schema[part]
            types = sorted(scheme.keys())
            makers = [make_maker(store, n, *scheme[n]) for n in types]
            setattr(self, part, namedtuple(part.capitalize(), types)(*makers))

        self.schema = schema

    @property
    def world(self):
        return self._world

    def set_world(self, world):
        if hasattr(world, 'name'):
            self._world = world.name
        else:
            self._world = world

    def get_shape(self, entry = None, index = 0):
        if hasattr(entry, 'shape'): # allow to pass entry as a Volume object
            entry = entry.shape

        return list_match(self.store.shapes.values(), entry, deref = lambda x: x.name)[index]



#!/usr/bin/env python
'''
Some low-level examples of builders.
'''

import gegede.builder
import gegede.construct

class SimpleBuilder(gegede.builder.Builder):
    '''A simple builder that makes a box and if it has a sub-builder,
    puts its volume at the center of the box.
    '''

    def configure(self, dx='1m', dy='1m', dz='1m', mat='Air', **kwds):
        #print 'Configuring "%s"' % self.name
        self.box_mat = mat      # assume made somewhere else
        self.box_dim = (dx,dy,dz)

    def construct(self, geom):
        #print 'Constructing "%s"' % self.name
        shape = geom.shapes.Box(self.name + '_box_shape', *self.box_dim)
        lv = geom.structure.Volume(self.name+'_volume', material=self.box_mat, shape=shape)
        self.add_volume(lv)

        # now hook up the sub-builder's LV if it exists
        if not self.builders:
            return
        subb = self.get_builder()
        if not subb.volumes:
            return

        sublv = subb.get_volume()
        p = geom.structure.Placement("%s_in_%s" % (sublv, lv.name), volume=sublv)
        # fixme: need to do something with this placement!
        return


def nested_boxes():
    '''
    Make some nested boxes by explicitly creating some builders
    '''
    blist = list()
    last_b = None
    sizes = ['1m', '100cm', '10cm', '1cm', '1mm']
    cfg = dict()

    for size in sizes:          # fake up some configuration and builder building
        sb = SimpleBuilder('box_of_size_%s' % size)
        cfg[sb.name] = dict(dx=size,dy=size,dz=size)
        blist.append(sb)
        if last_b:
            sb.builders[last_b.name] = last_b
        last_b = sb

    gegede.builder.configure(blist[0], cfg)
    geom = gegede.construct.Geometry()
    gegede.builder.construct(blist[0], geom)
    geom.set_world(blist[0].get_volume())
    return geom



#!/usr/bin/env python
'''
Some functions to help process the schema
'''
from collections import OrderedDict, namedtuple
from types import isquantity, toquantity

def make_converter(proto):
    '''Return a function that will convert its argument using <proto> as a
    prototype.  The <proto> is either a type or an instance of a Quantity
    or a string that can be parsed as a Quantity.
    '''
    if proto in (int, float):   # simple numerical type
        def topod(other):
            return proto(other)
        return topod

    if proto == str:            # treat string special to instrument some error checking
        def tostr(other):
            other = str(other)
            if other == "":
                raise ValueError, 'Empty string is an illegal value for str'
            if other is None:
                raise ValueError, 'None is an illegal value for str'
            return other
        return tostr

    if isquantity(proto):
        return toquantity(proto)

    # it better be of some type of self-converting prototype
    def toobj(other):
        ret = proto(other)
        return ret
    return toobj

def validate_input(proto, *args, **kwargs):
    '''Validate the input <args> and <kwargs> against the prototypes in <proto>.

    <proto> is a list of ("name",object) tuples.

    <args> is a list of positional arguments assumed to follow the
    ordering of <proto> but may be of shorter length.

    <kwargs> is a dictionary of "name"/object pairs supplying any
    remaining arguments.

    Any unspecified arguments are given the default from <proto> and
    no keys in <kwargs> may be given in addition to what are specified
    in <args>.

    A an ordered list of values are returned.
    '''
    #print 'VALIDATE INPUT', str(proto)

    members = OrderedDict()
    converters = dict()
    for name,pval in proto:
        c = make_converter(pval)
        converters[name] = c

        #print 'CONVERTER:',name,c,pval

        # set default
        if isquantity(pval):
            members[name] = c(pval)
        elif hasattr(pval, 'default'):
            #print 'SET DEFAULT:',name,pval.default
            members[name] = pval.default()
        else:
            members[name] = None

        # if isquantity(pval):
        #     pval = toquantity(pval)(pval)
        # members[name] = pval
        # converters[name] = make_converter(pval)

    already = list()
    for k,v in zip(members.keys(), args):
        members[k] = converters[k](v)
        #print 'ARGS:',k,v,members[k]
        already.append(k)

    for k,v in kwargs.items():
        if k not in converters.keys():
            raise ValueError, 'Object "%s" not in prototype' % (k,)
        if k in already:
            raise ValueError, 'Keyword argument already supplied as positional: %s' % k
        members[k] = converters[k](v)
        #print 'KWDS:',k,v,members[k]

    return members.values()
    

def make_maker(collector, ntname, *proto):

    '''
    Make a namedtuple class that does type-checking of its args against a prototype.

    The <proto> is a list of 2-tuples ("membername",<prototype object>)
    '''
    member_names = [p[0] for p in proto]
    def instantiator(objname, *args, **kwargs):
        if not objname:
            objname = "%s%06d" % (ntname, len(collector))

        if objname in collector:
            raise ValueError, 'Instance "%s" of type %s already in %s' % \
                (objname, ntname, collector)

        members = validate_input(proto, *args, **kwargs)

        NTT = namedtuple(ntname, ['name'] + member_names)
        obj = NTT(objname, *members)
        collector[objname] = obj
        #print 'INSTANTIATED:', obj
        return obj
    instantiator.__name__ = ntname
    instantiator.__doc__ = "%s: %s" % (ntname, ', '.join(member_names))
    return instantiator




#!/usr/bin/env python
'''
Export a gegede.construct.Geometry into a plain old data structure.
'''

# It would be nice to clean this code up to not hard-code assumptions
# about the data schema.

import pprint

def value2pod(val):
    if type(val) == tuple:
        return val
    return str(val)

def obj2pod(obj):
    d = {n:value2pod(v) for n,v in zip(obj._fields,obj)}
    return dict(classname = type(obj).__name__, **d)
                

def convert(geom):
    dat = dict()
    for section in geom.schema.keys():
        secdata = getattr(geom.store, section)
        things = [obj2pod(x) for x in secdata.values()]
        dat[section] = things
    return dat

def dumps(obj):
    return pprint.pformat(obj, indent=4) + '\n'

def output(obj, filename):
    fp = open(filename, 'w')
    fp.write('#!/usr/bin/env python\n')
    fp.write(dumps(obj))
    fp.close()



#!/usr/bin/env python
'''Some functions that do the interpretation of evaluated
configuration data into builder objects.
'''

def make_builder(dat, name = None):
    '''Return the a builder object of the given name from the
    configuration data <dat> and recursively call on any subbuilders.
    If no name is given return the first found.

    This consumes the 'class' and 'subuilders' entries of the configuration dat object.
    '''
    builder_objects = dict()
    if name is None:
        name = dat.items()[0][0]

    top = walk_builder_config_graph(dat, name, builder_objects)
    return top
    
def walk_builder_config_graph(dat, bname, builder_objects):
    '''Recur through the DAG of builder configs'''

    if bname in builder_objects:
        # been here.
        return builder_objects[bname]

    try:
        bdat = dat[bname]
    except KeyError,e:
        print 'No such builder configuration section: "bname"' % bname
        raise
    klass = bdat.pop('class')   # make exactly one instance
    bobj = klass(bname)
    builder_objects[bname] = bobj;

    subbuilders = bdat.pop('subbuilders',list())
    for sbname in subbuilders:
        sb = walk_builder_config_graph(dat, sbname, builder_objects)
        bobj.builders[sbname] = sb
    return bobj
    



#!/usr/bin/env python
'''
For guidance on writing GeGeDe exporter modules see:

  https://github.com/brettviren/gegede/blob/master/doc/exporting.org

'''

class Exporter(object):
    '''
    Wrapper around an exporter module which helps to enforce the conventions.
    '''
    def __init__(self, mod):
        if isinstance(mod, type("")):
            if not '.' in mod:
                mod = 'gegede.export.' + mod
            print 'Importing: "%s"' % mod
            exec('import %s' % mod)
            mod = eval(mod)
        self.mod = mod

        self.obj = None
        self.out = None

    def convert(self, geom):
        if self.obj:
            return self.obj
        self.obj = self.mod.convert(geom)
        return self.obj

    def dumps(self):
        if not self.obj:
            return None
        return self.mod.dumps(self.obj)

    def validate_object(self):
        if not hasattr(self.mod, 'validate_object'):
            return
        return self.mod.validate_object(self.obj)

    def output(self, filename):
        if self.out:
            return
        if not hasattr(self.mod, 'output'):
            print 'Warning: Module has no output() method: %s' % self.mod
            return
        self.out = filename
        self.mod.output(self.obj, filename)
        return

    def validate_output(self):
        if not hasattr(self.mod, 'validate_output'):
            return
        return self.mod.validate_output(self.obj, self.filename)

    pass
        



#!/usr/bin/env python
'''The module holds the default schema definition for the geometry description elements.

The schema is a dictionary with an item for each major category in the
schema.  The item's value is in a form which is category dependent.  

In general, leaf elements are described by a 2-tuple holding a data
member name and a prototypical value.  This value must be either an
object or a string that can be interpreted into an object in a
category-dependent manner.  

Shapes ("solids") follow the conventions shows in Geant4 and ROOT:

http://geant4.web.cern.ch/geant4/G4UsersDocuments/UsersGuides/ForApplicationDeveloper/html/Detector/geomSolids.html
http://root.cern.ch/root/html534/guides/users-guide/Geometry.html#shapes

 - the naming conventions of Geant4 are taken for the shapes (although we call them "shapes" not solids")

 - where applicable, the dimensions of solids are expressed as "half lengths" an such values are indicated with a "d" in their name (eg, "dx", "dy", "dz", "dphi").

 - all dimensions must be expressed in a quantity with the appropriate unit

'''

from types import Named, NameList, NamedTypedList

Schema = dict(

    shapes = dict(
        Box = (("dx","1m"), ("dy","1m"), ("dz","1m")),
        Tubs = (("rmin", "0m"), ("rmax","1m"), ("dz", "1m"),
               ("sphi","0deg"), ("dphi", "360deg")),
        Sphere = (("rmin", "0m"), ("rmax","1m"),
                    ("sphi","0deg"), ("dphi", "360deg"),
                    ("stheta","0deg"), ("dtheta", "180deg")),
        Boolean = (("type",str), ("first", Named), ("second", Named), 
                   ("pos", Named), ("rot", Named))

        # fixme: fill in the rest!
        ),

    matter = dict(
        Element = (("symbol",str), ("z",int), ("a","0.0g/mole")),
        Isotope = (("z",int), ("ia",int), ("a","0.0g/mole")),

        # A number of isotopes composed 
        Composition = (("symbol",str), ("isotopes", NamedTypedList(float))),

        # a material with no specific constituents
        Amalgam = (("z", int), ("a","0.0g/mole"), ("density", "0.0g/cc")),

        # A molecule is a Material with a number of elements
        Molecule = (("symbol",str), ("density", "0.0g/cc"), ("elements", NamedTypedList(int))),

        # A mixture is a Material that has a number of elements or
        # other materials added by mass fraction.
        Mixture = (("symbol",str), ("density", "0.0g/cc"), ("components", NamedTypedList(float))),

        # fixme, these need to also take state, temperature and pressure, radlen
        ),

    structure = dict(
        Position = (("x", "0m"), ("y","0m"), ("z","0m")),
        Rotation = (("x", "0deg"), ("y","0deg"), ("z","0deg")),
        Volume = (("material", Named), ("shape", Named), 
                  ("placements", NameList(str,0)),
                  ("params", NamedTypedList(str, 0))),
        Placement = (("volume", Named), ("pos", Named), ("rot", Named)),
    ),
)




#!/usr/bin/env python
'''
A (trivial) JSON representation of a gegede geometry
'''

import json
import pod

def convert(geom):
    return pod.convert(geom)

def dumps(obj):
    return json.dumps(obj, indent=2)

def output(obj, filename):
    open(filename,'w').write(dumps(obj))

def validate_object(obj):
    obj2 = json.loads(json.dumps(obj))
    assert obj == obj2

def validate_output(obj, filename):
    obj2 = json.load(filename)
    assert obj == obj2



#!/usr/bin/env python
'''
High level main functions.

The main runs in these stages:

 1) Generate
    - parse configuration
    - construct GeGeDe geometry objects
 2) Optional validation on GeGeDe objects
 3) Export
    - produce transient export-specific objects
    - optionally validate these objects
    - write these objects out to persistent storage
    - optionally validate the persistent representation

'''
import os


def generate(filenames, world_name = None):
    '''
    Return a geometry object generated from the given configuration file(s).
    '''
    import gegede.configuration
    import gegede.interp
    import gegede.builder
    import gegede.construct

    assert filenames
    cfg = gegede.configuration.configure(filenames)
    assert cfg
    wbuilder = gegede.interp.make_builder(cfg, world_name)
    gegede.builder.configure(wbuilder, cfg)
    geom = gegede.construct.Geometry()
    gegede.builder.construct(wbuilder, geom)
    assert len(wbuilder.volumes) == 1, 'Top level builder "%s" must only produce one LV, produced %d' % (wbuilder.name, len(wbuilder.volumes))
    geom.set_world(wbuilder.get_volume(0))
    print 'Generated world "%s"' % geom.world
    # fixme: here would be a good time to do some internal validation
    return geom


def main ():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-w", "--world", default=None,
                        help="World builder name")
    parser.add_argument("-f", "--format", default=None,
                        help = "Export format, guess by extension if not given")
    parser.add_argument("-o", "--output", default=None,
                        help="File to export to")
    parser.add_argument("-V", "--validate", action='store_true',
                        help="Validate geometry using built-in GeGeDe")
    parser.add_argument("-O", "--validate-object", action='store_true',
                        help="Validate exported, in-memory objects")
    parser.add_argument("-F", "--validate-file", action='store_true',
                        help="Validate exported, on-file objects")
    parser.add_argument("config", nargs='+',
                        help="Configuration file(s)")
    args = parser.parse_args()

    if not args.format and '.' not in args.output:
        raise parser.error("Can not guess format.  Need --format or --output with file extension")
    if not args.format:
        args.format = os.path.splitext(args.output)[1][1:]
    if args.output == '-':
        args.output = "/dev/stdout"

    geom = generate(args.config, args.world)
    if args.validate:
        import gegede.validation
        gegede.validation.validate(geom)
    from gegede.export import Exporter
    exporter = Exporter(args.format)
    print 'Converting with module: %s' % exporter.mod

    obj = exporter.convert(geom)
    if args.validate_object:
        exporter.validate_object()

    exporter.output(args.output)

    if args.validate_file:
        exporter.validate_file()
    return





from setuptools import setup


setup(name = 'gegede',
      version = '0.4',
      description = 'General Geometry Description',
      author = 'Brett Viren',
      author_email = 'brett.viren@gmail.com',
      license = 'GPLv2',
      url = 'http://github.com/brettviren/gegede',
      package_dir = {'':'python'},
      packages = ['gegede', 'gegede.schema', 'gegede.export',
                  'gegede.export.gdml', 'gegede.examples'],
      # These are just what were developed against.  Older versions may be okay.
      install_requires=[
          "pint >= 0.5.1",      # for units
          "lxml >= 3.3.5",      # for GDML export
      ],
      # implicitly depends on ROOT
      entry_points = {
          'console_scripts': [
              'gegede-cli = gegede.main:main',
              ]
      }
              
  )




#!/usr/bin/env python
'''
An export method producing for GDML.

http://lcgapp.cern.ch/project/simu/framework/GDML/doc/GDMLmanual.pdf
http://gdml1.web.cern.ch/gdml1/doc/g4gdml_howto.html
http://geant4.in2p3.fr/2007/prog/GiovanniSantin/GSantin_Geant4_Paris07_Materials_v08.pdf

http://lxml.de/tutorial.html
http://lxml.de/objectify.html
http://stackoverflow.com/questions/2850823/multiple-xml-namespaces-in-tag-with-lxml

'''

import os

schema_dir = os.path.join(os.path.dirname(__file__), 'schema')
schema_file = os.path.join(schema_dir, 'gdml.xsd')

from lxml import etree
from gegede import Quantity
from gegede.iter import ascending
from gegede.export import pod

def qtus(q,unit):
    'Quantity to unit string'
    return str(q.to(unit).magnitude)

def nt_qunit2xmldict(nt, unit):
    ret = dict(unit=unit)
    for k,v in zip(nt._fields,nt):
        if Quantity == type(v):
            v = str(v.to(unit).magnitude)
        else:
            v = str(v)
        ret[k] = v
    return ret

def D(obj):
    return str(obj.density.to('g/cc').magnitude)
def Atom(obj):
    return str(obj.a.to('g/mole').magnitude)
def Symbol(obj):
    return obj.symbol or ""

# def wash(geom):
#     '''Return a new geom object with its contents "washed" of any things
#     that will tickle GDML limitations.
#     '''
#     import gegede.construct
#     newg = gegede.construct.Geometry(geom.schema)
#     for store_name in geom.store._fields:
#         store = getattr(geom.store,store_name)
#         newstore = getattr(newg.store,store_name)
#         for name, obj in store.items():
#             newobj = type(obj)(obj.name.replace(' ','_'), *obj[1:])
#             newstore[name.replace(' ','_')] = newobj
#     return newg



def make_material_node(obj):
    '''
    Return an lxml.etree.Element made from the matter object.
    '''

    typename = type(obj).__name__
    node = None

    if typename == 'Element':
        node = etree.Element('element', name=obj.name, Z=str(obj.z), formula=Symbol(obj))
        node.append(etree.Element('atom', value=Atom(obj)))

    if typename == 'Isotope':
        node = etree.Element('isotope', name=obj.name, Z=str(obj.z), N=str(obj.ia))
        node.append(etree.Element('atom', type="A", value=Atom(obj)))

    if typename == 'Composition':
        node = etree.Element('element', name=obj.name)
        for isoname, isofrac in obj.isotopes:
            node.append(etree.Element('fraction', ref=isoname, n=str(isofrac)))

    if typename == 'Amalgam':
        node = etree.Element('material', name=obj.name, Z=float(obj.z))
        # fixme: units???
        node.append(etree.Element('D', value=D(obj)))
        node.append(etree.Element('atom', value=Atom(obj)))

    if typename == 'Molecule':
        node = etree.Element('material', name=obj.name, formula=Symbol(obj))
        node.append(etree.Element('D', value=D(obj)))
        for elename, elenum in obj.elements:
            node.append(etree.Element('composite', ref=elename, n=str(elenum)))


    if typename == 'Mixture':
        node = etree.Element('material', name=obj.name, formula=Symbol(obj))
        node.append(etree.Element('D', value=D(obj)))
        for compname, compfrac in obj.components:
            node.append(etree.Element('fraction', ref=compname, n=str(compfrac)))

    return node


def make_shape_node(shape):
    '''
    Return an lxml.etree.Element made from the <shape>
    '''
    lunit = 'cm'
    aunit = 'radian'
    typename = type(shape).__name__

    def unitify(unit, quant, scale=1.0):
        quant = quant * scale
        return str(quant.to(unit).magnitude)
    def dsize(quant):
        return unitify(lunit, quant, 2.0)
    def rsize(quant):
        return unitify(lunit, quant, 1.0)
    def ang(quant):
        return unitify(aunit, quant, 1.0)

    if typename == 'Box':
        dat = dict(name=shape.name, lunit=lunit,
                   x=dsize(shape.dx), y=dsize(shape.dy), z=dsize(shape.dz))
        return etree.Element('box', **dat)
                             
    if typename == 'Tubs':
        dat = dict(name=shape.name, lunit=lunit, aunit=aunit,
                   rmin=rsize(shape.rmin), rmax=rsize(shape.rmax), z=dsize(shape.dz),
                   startphi=ang(shape.sphi), deltaphi=ang(shape.dphi))
        return etree.Element('tube', **dat)

    if typename == 'Sphere':
        dat = dict(name=shape.name, lunit=lunit, aunit=aunit,
                   rmin=rsize(shape.rmin), rmax=rsize(shape.rmax),
                   startphi=ang(shape.sphi), deltaphi=ang(shape.dphi),
                   starttheta=ang(shape.stheta), deltatheta=ang(shape.dtheta))
        return etree.Element('sphere', **dat)

    if typename == 'Boolean':
        ele = etree.Element(shape.type, name=shape.name)
        # the rest are sub nodes.  why?  because, don't ask questions!
        ele.append(etree.Element('first', ref=shape.first))
        ele.append(etree.Element('second', ref=shape.second))
        ele.append(etree.Element('positionref', ref = shape.pos or 'center'))
        ele.append(etree.Element('rotationref', ref = shape.rot or 'identity'))
        return ele

    # etc....  Grow this as gegede.schema.Schema.shapes grows....

    return

def make_volume_node(vol, store):
    #print 'VOL',vol
    node_type = 'volume'
    if vol.material is None and vol.shape is None:
        node = etree.Element('assembly', name=vol.name)
    else:
        assert vol.material and vol.shape
        node = etree.Element('volume', name=vol.name)
        node.append(etree.Element('materialref', ref=vol.material))
        node.append(etree.Element('solidref', ref=vol.shape))
    for placename in vol.placements or []:
        place = store[placename]
        pvol = etree.Element('physvol')
        node.append(pvol)
        pvol.append(etree.Element('volumeref', ref=place.volume))
        if place.pos:
            pvol.append(etree.Element('positionref', ref=place.pos))
        else:
            pvol.append(etree.Element('positionref', ref='center'))
        if place.rot:
            pvol.append(etree.Element('rotationref', ref=place.rot))
        else:
            pvol.append(etree.Element('rotationref', ref='identity'))
    for parname, parval in vol.params or []:
        pnode = etree.Element('auxiliary', auxtype=parname, auxvalue=parval)
        node.append(pnode)
    return node


def convert(geom):
    '''
    Return an lxml.etree formed from the geometry
    '''

    gdml_node = etree.Element("gdml")

    # I have no idea what this means but it reproduces what other GDML
    # files have and w/out it Geant4 complains.
    gdml_node.set('{http://www.w3.org/2001/XMLSchema-instance}noNamespaceSchemaLocation',
                  'http://service-spi.web.cern.ch/service-spi/app/releases/GDML/schema/gdml.xsd')

    # <define>
    define_node = etree.Element('define')
    gdml_node.append(define_node)
    center = identity = None
    for name, obj in geom.store.structure.items():
        typename = type(obj).__name__.lower()
        node = None
        if typename == 'position':
            node = etree.Element('position', **nt_qunit2xmldict(obj, 'cm'))
            if obj.name == 'center':
                center = obj
        if typename == 'rotation':
            node = etree.Element('rotation', **nt_qunit2xmldict(obj, 'degree'))
            if obj.name == 'identity':
                identity = obj
        if node is not None:
            define_node.append(node)
        continue
    if center is None:
        define_node.append(etree.Element('position', name='center'))
    if identity is None:
        define_node.append(etree.Element('rotation', name='identity'))

    # <materials>
    materials_node = etree.Element('materials')
    gdml_node.append(materials_node)
    for name, obj in geom.store.matter.items():
        node = make_material_node(obj)
        if node is not None:
            materials_node.append(node)

    # <solids>
    solids_node = etree.Element('solids')
    gdml_node.append(solids_node)
    for obj in geom.store.shapes.values():
        node = make_shape_node(obj)
        if node is not None:
            solids_node.append(node)        
    

    # <structure>
    structure_node = etree.Element('structure')
    gdml_node.append(structure_node)
    for vol in ascending(geom.store.structure, geom.world):
        assert vol
        node = make_volume_node(vol, geom.store.structure)
        if node is not None:
            structure_node.append(node)

    # <setup>
    setup_node = etree.Element('setup', name="Default", version="0")
    gdml_node.append(setup_node)

    world_node = etree.Element('world', ref=geom.world)
    setup_node.append(world_node)

    return gdml_node


def validate(text):
    from StringIO import StringIO
    xsd_doc = etree.parse(schema_file)
    xsd = etree.XMLSchema(xsd_doc)
    xml = etree.parse(StringIO(text))
    okay = xsd.validate(xml)
    if not okay:
        print xsd.error_log
        raise ValueError, 'Invalid GDML'
    return True

def validate_object(obj):
    return validate(dumps(obj))

#def validate_output(obj, filename):
#    return False

def dumps(obj):
    '''
    Return a string representation of the object returned by convert.
    '''
    xml = etree.tostring(obj, pretty_print = True, xml_declaration = True)
    xml = xml.replace("'",'"')  # work around ROOT GDML import bug....
    # don't validate here
    return xml

def output(obj, filename):
    '''
    Save to file
    '''
    open(filename,'w').write(dumps(obj))



from gegede import construct

def airwaterboxes():
    g = construct.Geometry()
    h = g.matter.Element("Hydrogen","H",1,"1.01g/mole")
    n = g.matter.Element("Nitrogen", "N", 7, "14.01*g/mole")
    o = g.matter.Element("Oxygen", "O", 8, "16.0g/mole")
    water = g.matter.Molecule("Water", density="1.0kg/l", elements=(("Oxygen",1),("Hydrogen",2)))
    air = g.matter.Mixture("Air", density = "1.290*mg/cc", 
                           components = (("Nitrogen", 0.7), ("Oxygen",0.3)))

    U235 = g.matter.Isotope("U235", z=92, ia=235, a="235.0439242 g/mole")
    U238 = g.matter.Isotope("U238", z=92, ia=238, a="238.0507847 g/mole")
    enriched_uranium = g.matter.Composition("enriched_U", symbol="U", 
                                            isotopes=(("U235",0.8), ("U238",0.2)))

    box1 = g.shapes.Box("box1",'1cm','2cm','3cm')    
    box2 = g.shapes.Box("box2",'1m','2m','3m')    
    pos = g.structure.Position(None, '1cm',z='2cm')
    rot = g.structure.Rotation('', x='90deg')
    lv1 = g.structure.Volume('a_box', material=water, shape=box1)
    lv1inlv2 = g.structure.Placement("lv1_in_lv2", volume=lv1, pos=pos, rot=rot)
    lv2 = g.structure.Volume('the_world', material = air, shape=box2,
                             placements = [lv1inlv2], params= (("foo",42), ("bar","baz")))
    g.set_world(lv2)
    return g




#!/usr/bin/env python
'''
A GGD builder
'''

from collections import OrderedDict

from .util import list_match

class Builder(object):
    '''
    A Builder base class

    This class maintains the following data members:

    .builders - an ordered dictionary of sub-builder objects

    .volumes - an ordered dictionary of top-level volume objects

    For adding to these, see self.add_* methods named after these members
    '''


    def __init__(self, name):
        '''
        Create an instance of the builder with the given <name>.

        The <name> is used to later locate the builders configuration.
        '''
        self.name = name
        self.builders = OrderedDict()
        self.volumes = OrderedDict()
        
    
    def get_builders(self, entry = None):
        '''
        Return sequence of builder objects that match the entry.
        '''
        return list_match(self.builders.values(), entry, deref = lambda x: x.name)

    def get_builder(self, entry = None, index = 0):
        '''
        Return the one builder that matches the entry.
        '''
        b = list_match(self.builders.values(), entry, deref = lambda x: x.name)[index]
        #print 'get_builder("%s") -> %s:%s' % (entry, b, b.name)
        return b

    def get_volumes(self, entry = None):
        '''
        Return sequence of volume objects that match the entry.
        '''
        return list_match(self.volumes.values(), entry, deref = lambda x: x.name)

    def get_volume(self, entry = None, index = 0):
        '''
        Return the one volume that matches the entry.
        '''
        return list_match(self.volumes.values(), entry, deref = lambda x: x.name)[index]

    def add_volume(self, *vols):
        '''Register top-level logical volumes to self.volumes.  

        <vols> is a list logical volume objects.

        No return value.

        Subclass should call this in construct().
        '''

        for v in vols:
            if v.name in self.volumes:
                print 'Volume already exists: "%s" out of %d volumes' % (v.name, len(self.volumes))
                print '\n'.join(self.volumes)
                continue
            self.volumes[v.name] = v
        return

    def add_builder(self, name, klass):
        '''Add a builder of type <klass> and given name <name>.

        No return value.

        klass may be an instance of a Builder class or a string
        holding the fully-qualified name of a class.
        '''
        if self.get_builders(name):
            print ('Builder already registered: "%s"' % name)
            return
            
        if (isinstance(klass, type(""))):
            mod,name = klass.rsplit('.',1) # this seems dirty
            exec ('import %s' % mod)
            klass = eval(klass)
        builder = klass(name)
        self.builders[name] = builder
        return

    def configure(self, **kwds):
        '''Process my configuration.

        The <kwds> dictionary holds any configuration parameters
        relevant to this builder.

        No return value.

        Subclass may override.

        If a self.defaults value exists it updated with kwds and the
        items are set as data members of self.  

        The builder may add sub-builder objects inside this method.
        See self.add_builder().  They will have their own configure()
        method called after this method exits.

        '''
        if hasattr(self, 'defaults'):
            if not set(kwds).issubset(self.defaults): # no unknown keywords
                msg = 'Unknown parameter in: "%s"' % (', '.join(sorted(kwds.keys())), )
                raise ValueError,msg
            self.__dict__.update(**self.defaults)    # stash them as data members
            self.__dict__.update(**kwds)             # and update any from user

        return

    
    def construct(self, geom):
        '''Construct the builder's geometry.

        The <geom> object is an instance of gegede.construct.Geometry.

        Sub class should override.

        Implementation should produce all materials and logical or
        physical volumes needed.  Any sub-builders will have their
        construct() method called prior to this one and so will have
        their .volumes ordered dictionary populated.

        '''
        return


# Functions to manage builder hierarchy:

def configure(builder, cfg):
    '''Configure a builder with its item in the <cfg> dictionary and
    recurse through any sub-builders.
    '''

    if hasattr(builder, '_configured'): 
        return                  # only configure once
    builder.configure(**cfg.get(builder.name, dict()))
    builder._configured = True;

    # recurs on child builders
    for other in builder.builders.values():
        configure(other, cfg)
    return

def construct(builder, geom):
    '''Call the construct method on the builder after recursing any
    sub-builders'''
    for other in builder.builders.values():
        construct(other, geom)

    if hasattr(builder, '_constructed'):
        return
    builder.construct(geom)
    builder._constructed = True
    return


    



#!/usr/bin/env python
'''
'''

from gegede import Quantity

def wash_units(obj, lunit='cm', aunit='radian'):
    '''Return a tuple (newobj,used) where <newobj> is <obj> with any
    Quantities placed into explicit units given by lunit for length
    and aunit for angle.  The <used> is a dict of all units used.
    '''

    ret = dict()
    used = dict()
    for k,v in zip(obj._fields, obj):
        if type(v) == Quantity:
            if 'meter' in v.to_base_units().units:
                used['lunit'] = lunit
                ret[k] = v.to(lunit).magnitude
                continue
            if 'radian' in v.to_base_units().units:
                used['aunit'] = aunit
                ret[k] = v.to(aunit).magnitude
                continue
            pass
        ret[k] = v
    newobj = type(obj)(**ret)
    return (newobj,used)

import re
def list_match(values, entry = None, deref = lambda x: x):
    '''
    If not entry, return values.  

    If entry is int, return values at entry (at most single element list)

    If entry is a callable, call it on deref(value) and return value if callable returns true.

    O.w. assume str and first try exact match otherwise interpret as regexp and return values matching deref(value).
    '''
    #print "list_match(%s,%s)" % (', '.join(['%s:%s' % (type(v),v.name) for v in values]), entry)

    if entry is None:
        return values

    if isinstance(entry, int):
        return [values[entry]]

    if callable(entry):
        ret = list()
        for v in values:
            if entry(deref(v)):
                ret.append(v)
        return ret

    # assume string, first as exact match
    ret = list()
    for v in values:
        if entry == deref(v):
            ret.append(v)
    if ret:
        return ret
    

    # then as regex
    rx = re.compile(entry, re.I)
    for v in values:
        if rx.search(deref(v)) is None:
            continue
        ret.append(v)
    return ret




#!/usr/bin/env python
'''
gegede configuration
'''
import os
import re
from collections import OrderedDict

def parse(filenames):
    '''Parse configuration files.

    Return the parser object.'''

    try:                from ConfigParser import SafeConfigParser
    except ImportError: from configparser import SafeConfigParser
    cfg = SafeConfigParser()
    cfg.optionxform = str       # want case sensitive
    
    if isinstance(filenames, type("")):
        filenames = [filenames]
    for fname in filenames:
        if not os.path.exists(fname):
            raise ValueError, 'No such file: %s' % fname
        cfg.read(fname)
    return cfg


def cfg2pod(cfg):
    '''
    Convert a ConfigParser object to a plain-old-data structure
    '''
    pod = OrderedDict()
    for secname in cfg.sections():
        secdat = OrderedDict()
        for k,v in cfg.items(secname):
            secdat[k] = v
        pod[secname] = secdat
    return pod

interp_reobj = re.compile(r'{([\w:]+)}')
def interpolate_value(value, secname, sections):
    def getter(match):
        m = match.group(1)
        try:
            name, key = m.split(':')
        except ValueError:
            key = m
            name = secname
        return sections[name][key]
    ret = re.subn(interp_reobj, getter, value)
    return ret[0]

def interpolate(sections):
    '''Perform string interpolation values <sections>.

    The <sections> data structure is a dictionary keyed by SECTION name.

    Each section value is a dictionary keyed by KEY with string values.

    Interpolation is performed on literals matching {KEY} or {SECTION:KEY}.
    '''
    while True:
        changed = False
        for secname, secdict in sections.items():
            for key, val in secdict.items():
                newval = interpolate_value(val, secname, sections)
                if newval == val:
                    continue
                changed = True
                sections[secname][key] = newval
        if changed:
            continue
        break
        


def make_class(fqclass):
    mod, cls = fqclass.rsplit('.',1)
    exec('import %s' % mod) # better way?
    return eval(fqclass)

def make_value(v, **kwds):
    from . import Quantity
    return eval(v, globals(), dict(Q=Quantity, **kwds))


def evaluate(pod):
    '''
    Evaluate and replace each value.
    '''
    ret = OrderedDict()
    for secname, secdat in pod.items():
        newdat = OrderedDict()
        for k,v in secdat.items():
            if k == 'class':
                newdat[k] = make_class(v)
                continue
            newdat[k] = make_value(v, **newdat)
        ret[secname] = newdat
    return ret

def configure(filenames):
    '''
    Return evaluated configuration 
    '''
    cfg = parse(filenames)
    assert cfg.sections()
    pod = cfg2pod(cfg)
    assert pod

    interpolate(pod)

    dat = evaluate(pod)
    assert dat
    return dat

